# -*- coding: utf-8 -*-
# Generated by Django 1.10.8 on 2018-03-30 21:08
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('seeker', '0004_auto_20180322_1412'),
    ]

    operations = [
        migrations.AlterField(
            model_name='advancedsavedsearch',
            name='search_object',
            field=models.TextField(),
        ),
    ]
