{% load i18n %}

{{ form.non_field_errors }}
<div class="form-group">
    <label for="{{ form.name.id_for_label }}">{% trans "Name" %}</label>
    {{ form.name }}
    {{ form.name.errors }}
</div>
<div class="form-group">
    <label class="checkbox-inline">
        {{ form.default }}
        {% trans "Make this your default search for this page" %}
        {{ form.default.errors }}
    </label>
</div>
{{ form.url.as_hidden }}
{{ form.search_object.as_hidden }}
