# -*- coding: utf-8 -*-
# Generated by Django 1.10.8 on 2018-03-22 18:12
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('seeker', '0003_auto_20180214_1501'),
    ]

    operations = [
        migrations.CreateModel(
            name='AdvancedSavedSearch',
            fields=[
                ('savedsearch_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='seeker.SavedSearch')),
                ('search_object', models.TextField(blank=True)),
            ],
            bases=('seeker.savedsearch',),
        ),
        migrations.RemoveField(
            model_name='savedsearch',
            name='data',
        ),
        migrations.RemoveField(
            model_name='savedsearch',
            name='saved',
        ),
    ]
