# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-06-25 15:39
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('seeker', '0005_auto_20180330_1708'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='advancedsavedsearch',
            options={'ordering': ('name',), 'verbose_name': 'Advanced Saved Search', 'verbose_name_plural': 'Advanced Saved Searches'},
        ),
        migrations.AlterModelOptions(
            name='savedsearch',
            options={'ordering': ('name',), 'verbose_name': 'Saved Search', 'verbose_name_plural': 'Saved Searches'},
        ),
    ]
